#!/bin/bash
# ======================================================================
#
# MULTI REMOTE EXECUTION
#
# ----------------------------------------------------------------------
# 2023-03-13       <axel.hahn@unibe.ch>  WIP
# 2023-03-24  0.3  <axel.hahn@unibe.ch>  show last execution
# 2023-04-17  0.4  <axel.hahn@unibe.ch>  add ssh user; fix display of last execution
# 2023-06-05  0.5  <axel.hahn@unibe.ch>  fix missing -e param
# 2023-09-07  0.6  <axel.hahn@unibe.ch>  execute each command in a single terminal
# 2023-09-08  0.7  <axel.hahn@unibe.ch>  show if a command is running; use color.class.sh
# 2023-09-08  0.8  <axel.hahn@unibe.ch>  fix ERROR: color 'output' is not a name nor a value (...)
# 2023-10-06  0.9  <axel.hahn@unibe.ch>  better matching of hostname in processes and log
# 2023-10-13  0.10 <axel.hahn@unibe.ch>  change log format; simplify command string
# 2023-10-16  0.11 <axel.hahn@unibe.ch>  change date format; serverlist timestamp hides date of current day
# 2023-10-16  0.12 <axel.hahn@unibe.ch>  fix last success value for long hostnames
# 2023-10-18  0.13 <axel.hahn@unibe.ch>  rewrite: config handled by json; filtering by tags
# 2023-10-18  0.14 <axel.hahn@unibe.ch>  automatically init of userconfig; check terminal
# 2023-10-19  0.15 <axel.hahn@unibe.ch>  Fix: Show running terminal in server list
# 2024-05-15  0.16 <axel.hahn@unibe.ch>  add profiles; rename internal variables
# ======================================================================

_version=0.16
_self=$( basename $0 )
cd "$( dirname $0 )"

# file definitions
MRE_CONFIGFILE=config/userconfig.sh

# profile (folder below ./config/)
MRE_PROFILE=default
MRE_JSONFILE=
MRE_LOGFILE=logs/returncodes.log

MRE_SSHUSER=ladmin
MRE_COMMAND="hostname -f" 
MRE_TERMINAL=gnome-terminal
MRE_TAGS=

# default colors
COLOR_PRESET_cmd=("blue")
COLOR_PRESET_bastion=("purple")
COLOR_PRESET_lastexec=("green")
COLOR_PRESET_terminal=("brown")
COLOR_PRESET_prompt=("green")
COLOR_PRESET_output=("darkgray")
COLOR_PRESET_ok=("green")
COLOR_PRESET_error=("red")

. color.class.sh

. "${MRE_CONFIGFILE}" 2>/dev/null

USAGE="
SYNTAX:
    $_self [OPTIONS] [SERVERLIST]

OPTIONS:
    -c|--command [COMMAMD]
                 set a command for remote execution. A 2nd value
                 for the new command is optional. Without param
                 you reach an interactive mode.
                 Save config and exit.
    -h|--help    show this help and exit
    -l|--list    list servers
    -p|--profile [PROFILE]
                 set a profile for network environment
    -t|--tag [TAGNAME(s)]
                 set a tag to filter serverlist.
                 To set multiple tags quote it as a single string
                 and seperate them wih a space. Without param
                 you reach an interactive mode.
                 Save config and exit. 
PARAMETERS:
    The serverlist is a combination of servers that should
    execute the command [$MRE_COMMAND].
    It can be the number of a server or a 
    A special item is ALL - it executes the command on all servers

    Without a parameter you reach the interactive mode.

EXAMPLES:
    $_self -c               - set a new command. Enter it on the prompt
    $_self -c \"sudo reboot\" - set a given command
    $_self -t \"mon live\"    - set filter tags \"mon\" and \"live\"
    $_self -l               - list servers that match all tags (AND condition)
    $_self 1 2 5            - run the command on 1st, 2nd and 5th server
    $_self monitor          - run the command on the first server that matches 'monitor'
"

# ----------------------------------------------------------------------
# FUNCTIONS
# ----------------------------------------------------------------------

# ---------- PROFILE

# set a profile; switches to interactive mode if no param is given
# param  string  optional: new profile
function _setProfile(){
    local new="$1"
    if [ -z "$new" ]; then
        find config -maxdepth 1 -type d | sed "s#^config[/]*##g"
        echo
        echo -n "Current: "
        color.echo "cmd" "$MRE_PROFILE"
        echo "Enter a profile name"
        color.print "prompt" "> "
        read -r new
    fi
    if [ -n "$new" ]; then
        MRE_PROFILE=$new
        _save
        _getProfile
    else
        echo "Doing nothing."
    fi
}


# check current profilename; if correct then set MRE_JSONFILE that
# points to a json with the serverlist
function _getProfile(){
    if [ -z "$MRE_PROFILE" ]; then
        echo "WARNING: profile is empty - set it to [default]"
        MRE_PROFILE=default
    fi
    if [ ! -d "config/${MRE_PROFILE}/" ]; then
        echo "ERROR: profile [${MRE_PROFILE}] is invalid - ./config/${MRE_PROFILE}/ does not exist"
        exit 1
    fi
    if [ ! -f "config/${MRE_PROFILE}/serverlist.json" ]; then
        echo
        echo "WELCOME!"
        echo "We need a file <serverlist.json> to start."
        echo
        echo "Create a file config/${MRE_PROFILE}/serverlist.json ... OR"
        echo "You can create a new profile directory ./config/<PROFILE>/ and for a"
        echo "seperate network with custom servers and commands."
        echo "HINT: Copy the template config/default/serverlist.json.dist to start."
        echo
        echo "If you want to switch to another profile use $(basename $0 ) -p "
        echo
        exit 1
    fi
    MRE_JSONFILE=config/${MRE_PROFILE}/serverlist.json
}

# ---------- COMMANDS

# show a list of stored commands
function _getCommandlist(){
    echo "---- List default commands :: ./config/commands.txt"
    echo
    color.preset "output"
    cat "./config/commands.txt"
    color.reset
    if [ -f "./config/${MRE_PROFILE}/commands.txt" ]; then
        echo
        echo "---- Profile specific commands :: ./config/${MRE_PROFILE}/commands.txt"
        echo
        color.preset "output"
        cat "./config/${MRE_PROFILE}/commands.txt"
        color.reset
    fi
}

# set a command; switches to interactive mode if no param is given
# param  string  optional: new command
function _setCommand(){
    local newcommand="$1"
    if [ -z "$newcommand" ]; then
        _getCommandlist
        echo
        echo -n "Current: "
        color.echo "cmd" "$MRE_COMMAND"
        echo "Enter a command to execute (from given list or a new one)"
        color.print "prompt" "> "
        read -r newcommand
    fi
    if [ -n "$newcommand" ]; then
        MRE_COMMAND=$newcommand
        _save
    else
        echo "Doing nothing."
    fi
}

# ---------- Json related: maintenance, servers, tags

# get json data; this functions removes comments befind "#"
# global: $MRE_JSONFILE
function _getJSON(){
    cat $MRE_JSONFILE | sed "s,#.*$,,g"
}

# get a list of maintenance tags
# param1  string  value of matching tag 1
# paramN  string  value of matching tag N
function _getMaintenances(){
    _getJSON | jq ".maintenance[] .label "
}

# get a list of server entries matching the given tags
# Remark: a tag with spaces will be splittet
# param1  string  value of matching tag 1
# paramN  string  value of matching tag N
function _getServerJson(){
    local _cmd=".servers[]"

    for mytag in $*
    do
        _cmd+=' | select(.tags[] == "'$mytag'")'
    done
    _getJSON | jq "${_cmd} | { host: .host, bastion: .bastion, tags: .tags } "
}

# get a single host record with all its properties as JSON
# param  string  hostname
function _getHostData(){
    local _host="$1"
    _getServerJson "$MRE_TAGS" | jq ". | select ( .host == \"$_host\" ) "
}

# get a property from a host item.
# the given hostname must match a host that matches given tags
# param  string  hostname
# param  string  property key; one of host|bastion|tags
function _getHostProperty(){
    local _host="$1"
    local _property="$2"

    _getHostData "$myserver" | jq ".${_property}" | tr -d '"' | sed "s#^null\$##"
}

# show a list of all tags on screen
function _gettags(){
    local _spacer="     "

    echo "-- all current tags in hosts:"
    color.preset output
    _getJSON | jq  ".servers[] | .tags " | grep '"' | tr -d ',"' | sort -u | sed "s#^ *##g" | sed "s#^#${_spacer}#g"
    color.reset

    echo "-- maintenance tags:"
    color.preset output
    _getMaintenances | sed "s#^#${_spacer}#g" | tr -d '"'
    color.reset
}

# set a command; switches to interactive mode if no param is given
# param  string  optional: new command
function _setTags(){
    local newtags="$1"
    if [ -z "$newtags" ]; then
        _gettags
        echo
        echo -n "Current: "
        color.echo "cmd" "$MRE_TAGS"
        echo "Enter tags to filter hosts; seprate mulitple tags with space (from given list or a new one)"
        color.print "prompt" "> "
        read -r newtags
    fi
    if [ -n "$newtags" ]; then
        MRE_TAGS="$newtags"
        _save
    else
        echo "Doing nothing."
    fi
}

# ---------- save

# save params into user config
# gobal  MRE_CONFIGFILE  string  filename of user config file
function _save(){
    cat << EOC > "${MRE_CONFIGFILE}"
#!/bin/bash
# CONFIG for $_self - saved $(date)

# profile
export MRE_PROFILE=$MRE_PROFILE

# admin user to connecto to other hosts
export MRE_SSHUSER=$MRE_SSHUSER

# terminal program to open per connection
export MRE_TERMINAL=$MRE_TERMINAL

# filter hosts, set command
export MRE_TAGS="$MRE_TAGS"

# command to execute on target host(s)
export MRE_COMMAND="$MRE_COMMAND"

# colors
export COLOR_PRESET_cmd=( ${COLOR_PRESET_cmd[*]} )
export COLOR_PRESET_bastion=( ${COLOR_PRESET_bastion[*]} )
export COLOR_PRESET_lastexec=( ${COLOR_PRESET_lastexec[*]} )
export COLOR_PRESET_terminal=( ${COLOR_PRESET_terminal[*]} )
export COLOR_PRESET_prompt=( ${COLOR_PRESET_prompt[*]} )
export COLOR_PRESET_output=( ${COLOR_PRESET_output[*]} )
export COLOR_PRESET_ok=( ${COLOR_PRESET_ok[*]} )
export COLOR_PRESET_error=( ${COLOR_PRESET_error[*]} )
EOC
    echo "saved:"
    ls -l "${MRE_CONFIGFILE}"
}

# ---------- Execute commands

# get my sub processes in open terminal windows
function _getTerminals(){
    ps -ef | grep ssh | grep "$MRE_COMMAND"
}

# get my sub processes in open terminal windows
# param  string  process list
function _isOpen(){
    local _processes="${1}"
    local _srv="${2}"
    local _bst="${3}"
    local _openterm
    local _running

    _running=

    # for checking the bastion itself: remove processes of ssh tunnels
    test -z "$_bst" && _processes=$( grep -v 'ssh -W %h:%p' <<< "$_processes" )

    # gnome-terminal -e ... does not appear in process list
    # _openterm=$( grep "@$_srv" <<< "$_processes" | grep "$MRE_TERMINAL") 
    _openterm=$( grep "@$_srv" <<< "$_processes") 
    if [ -n "$_openterm" ]; then
        _openterm="TERM"
        _running=$( grep "$_srv" <<< "$_processes" | grep -v "$MRE_TERMINAL" | grep -v "bash -c")
        test -n "$_running" && _openterm+=" >> RUNNING"
        # test -n "$_running" || _openterm+=" <done>"
        echo "$(color.preset terminal )$_openterm$(color.reset)"
    fi

}


# helper: get a numbered list of servers to connect to 
function showServers(){

    local _openterm
    local _running
    local _processes

    _processes=$( _getTerminals )

    # --- show bastion and last successful execution    
    echo "---- List of servers - $( date )"
    echo

    color.underline
    echo "        Server                                    Bastion (20 chars)           last success  Running"
    color.reset

    _getServerJson "$MRE_TAGS" | jq -c | while read -r line
    do
        local _srv=$( echo "$line" | jq ".host"    | tr -d '"')
        local _bst=$( echo "$line" | jq ".bastion" | tr -d '"' | cut -c 1-20 | sed "s#^null\$##g")

        if [ -f "$MRE_LOGFILE" ]; then
            _lastexec=$( cat "$MRE_LOGFILE" | grep "|host=$_srv|rc=0|" | grep -F "$MRE_COMMAND" | tail -1 | cut -f 1 -d "|" | sed "s#$( date +%Y-%m-%d )##g")
        else
            _lastexec=
        fi
        _openterm=$( _isOpen "$_processes" "$_srv" "$_bst")
        _srv=$( echo "$_srv" | cut -c 1-40 )

        printf "%-40s  %-20s  %19s  %s\n" "${_srv}" "${_bst}" "${_lastexec}" "${_openterm}"
    done | nl

}

# helper: get the first matching server based on param which can be a 
# number or part of a fqdn
# param  integer|string  number or part of servername to search
function getServerByNumber(){
    local _searchtext="$1"

    # generate a list with number and hostname
    local _servers; _servers=$( _getServerJson "$MRE_TAGS" | jq ".host" -c | tr -d '"' | nl | awk '{ print $1 " " $2 }' )
    (
        # match a given number first ... and then in hostnames
        grep "^${_searchtext} " <<< "$_servers" \
            || grep "${_searchtext}" <<< "$_servers"
    ) | head -1 | cut -f 2 -d " "
}

# run the command on a target system
# param  integer|string  number or part of servername to search
function run(){
    local myserver="$1"
    
    # --- deny multiple instances
    _processes=$( _getTerminals )
    _openterm=$( _isOpen "$_processes" "$myserver")
    if [ -n "$_openterm" ]; then
        # echo
        # grep "$myserver" <<< "$_processes" | grep "$MRE_TERMINAL"
        echo
        echo "ABORT: $myserver has an open terminal already: $_openterm"
        echo
        sleep 0.5
        return 1
    fi

    # --- create command line
    bastion=$( _getHostProperty "$1" "bastion" )

    remoteCmd=
    remoteCmd+="ssh "
    test -n "$bastion" && remoteCmd+='-o ProxyCommand="ssh -W %h:%p -q '${MRE_SSHUSER}'@'$bastion'" '
    remoteCmd+="${MRE_SSHUSER}@${myserver} "
    remoteCmd+=' "'$MRE_COMMAND'"'

    command="
        echo \$( date ) START $remoteCmd ...
        echo _______________________________________________________________________________
        echo
        $remoteCmd
        rc=\$?
        echo _______________________________________________________________________________
        echo \$( date +\"%Y-%m-%d %H:%M:%S\" )\|host=$myserver\|rc=\$rc\|command=$remoteCmd >> $MRE_LOGFILE
        echo \$( date ) END $remoteCmd
        test \$rc -eq 0 && ( echo -en \"\e[32m>>> OK >>> \"              )
        test \$rc -eq 0 || ( echo -en \"\e[31m>>> ERROR: rc =\$rc >>> \" )
        echo -en \"\e[34m  Press RETURN to exit\"
        read 
    "

    # --- open terminal
    color.preset "output"
    set -vx
    $MRE_TERMINAL -e " bash -c '$command' " &
    set +vx
    color.reset
}

# loop over list of servers and run the command
# param  string  list of servers to handle (seperated by space)
function looper(){
    for mysrv in $*
    do
        myserver=$( getServerByNumber "$mysrv")
        run "$myserver"
        shift
    done
}

# ----------------------------------------------------------------------
# MAIN
# ----------------------------------------------------------------------

cat <<EOH
_______________________________________________________________________________

 MULTI REMOTE EXECUTION WITH SSH                                         ______
________________________________________________________________________/ v$_version
EOH

test ! -f "${MRE_CONFIGFILE}" && ( echo ; echo "INIT: Creating user config..."; _save )
test ! -f "./config/commands.txt" && ( echo ; echo "INIT: Creating default commands ./config/commands.txt ..."; cp "./config/commands.txt.dist" "./config/commands.txt" )

# --------- param check

if echo "$*" | grep "\-h" >/dev/null; then
    echo "$USAGE"
    exit 0
fi

# --------- check terminal program

if ! which $MRE_TERMINAL >/dev/null 2>&1; then
    color.echo "error" "ERROR: the terminal [$MRE_TERMINAL] was not found."
    echo
    echo "In the [${MRE_CONFIGFILE}] set a line:"
    color.echo "output" "MRE_TERMINAL=[your-terminal-program-here]"
    echo
    echo "Ensure that it supports the parameter -e to execute a command."
    echo "These tools are known to be OK: konsole, gnome-terminal, xterm"
    echo
    exit 1
fi

if ! which jq >/dev/null 2>&1; then
    color.echo "error" "ERROR: the JSON parser [jq] was not found."
    echo
    exit 1
fi

# --------- param check II

serverlist=

while [[ "$#" -gt 0 ]]; do case $1 in
    -c|--command)   _setCommand "$2"; exit 0;;
    -h|--help)      echo "$USAGE"; exit 0;;
    -l|--list)      showServers ; echo; exit 0;;
    -t|--tag)       _setTags "$2"; exit 0;;
    -p|--profile)   _setProfile "$2"; exit 0;;
    *) if echo "$1" | grep "^-" >/dev/null; then
            color.echo "error" "ERROR: Unknown parameter: $1"; echo "${USAGE}"; exit 1;
        else
            serverlist+="$1 "; shift
        fi
        ;;
esac; done

# --------- Check JSON

_getProfile

if ! _getJSON | jq >/dev/null; then
    color.echo "error" "ERROR: The JSON data seem to be corrupt - fix $MRE_JSONFILE"
    exit 2
fi
if _getServerJson 2>& 1| grep -F "error (at"; then
    color.echo "error" "ERROR: The JSON data seem to be corrupt - fix $MRE_JSONFILE"
    exit 2
fi


if echo "$serverlist" | grep "ALL" >/dev/null; then
    echo "INFO: Keyword 'ALL' detected ... looping over all servers"
    serverlist=$( _getServerlist | cut -f 1 -d " " )
fi

if [ -n "$serverlist" ]; then
    # execute the current command on given server(s)
    looper "$serverlist"
    echo "DONE"
    exit 0
fi

# --------- interactive mode

echo
while true; do

    showServers
    echo
    echo "Special commands:"
    echo -n " ':c' change command "; color.echo "cmd" "$MRE_COMMAND"
    echo -n " ':t' change tags    "; color.echo "cmd" "$MRE_TAGS"
    echo -n " ':p' change profile "; color.echo "cmd" "$MRE_PROFILE"
    echo "Enter numbers or parts of servernames. Use space as divider."
    echo "Just press return to update server infos amd running status."

    color.print "prompt" "> "
    # read -t 60 -r srvnumber
    read -r srvnumber

    if [ "$srvnumber" = ":c" ]; then
        _setCommand
    elif [ "$srvnumber" = ":t" ]; then
        _setTags
    elif [ "$srvnumber" = ":p" ]; then
        _setProfile
    else
        # execute the current command on given server(s)
        looper "$srvnumber"
    fi
    echo
done

# ----------------------------------------------------------------------
